﻿using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Runtime.InteropServices;
using System.Windows.Forms;

namespace TestProject
{
    /// <summary>
    /// WEBP 헬퍼
    /// </summary>
    public sealed class WEBPHelper : IDisposable
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// WEBP 최대 차원
        /// </summary>
        private const int WEBP_MAXIMUM_DIMENSION = 16383;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 정보 구하기 - GetInformation(sourceByteArray, width, height, hasAlpha, hasAnimation, format)

        /// <summary>
        /// 정보 구하기
        /// </summary>
        /// <param name="sourceByteArray">소스 바이트 배열</param>
        /// <param name="width">너비</param>
        /// <param name="height">높이</param>
        /// <param name="hasAlpha">알파 여부</param>
        /// <param name="hasAnimation">애니메이션 여부</param>
        /// <param name="format">포맷</param>
        public void GetInformation
        (
            byte[]     sourceByteArray,
            out int    width,
            out int    height,
            out bool   hasAlpha,
            out bool   hasAnimation,
            out string format
        )
        {
            VP8StatusCode statusCode;
            GCHandle      sourceHandle = GCHandle.Alloc(sourceByteArray, GCHandleType.Pinned);

            try
            {
                IntPtr pinnedSourceHandle = sourceHandle.AddrOfPinnedObject();

                WEBPBitStreamFeatures features = new WEBPBitStreamFeatures();

                statusCode = WIN32Helper.WEBPGetFeatures(pinnedSourceHandle, sourceByteArray.Length, ref features);

                if(statusCode != 0)
                {
                    throw new Exception(statusCode.ToString());
                }

                width  = features.Width;
                height = features.Height;

                if(features.HasAlpha == 1)
                {
                    hasAlpha = true;
                }
                else
                {
                    hasAlpha = false;
                }

                if(features.HasAnimation == 1)
                {
                    hasAnimation = true;
                }
                else
                {
                    hasAnimation = false;
                }

                switch(features.Format)
                {
                    case 1  : format = "lossy";     break;
                    case 2  : format = "lossless";  break;
                    default : format = "undefined"; break;
                }
            }
            finally
            {
                if(sourceHandle.IsAllocated)
                {
                    sourceHandle.Free();
                }
            }
        }

        #endregion
        #region 디코딩하기 - Decode(sourceByteArray)

        /// <summary>
        /// 디코딩하기
        /// </summary>
        /// <param name="sourceByteArray">소스 바이트 배열</param>
        /// <returns>비트맵</returns>
        public Bitmap Decode(byte[] sourceByteArray)
        {
            Bitmap     targetBitmap     = null;
            BitmapData targetBitmapData = null;
            GCHandle   sourceHandle     = GCHandle.Alloc(sourceByteArray, GCHandleType.Pinned);

            try
            {
                GetInformation
                (
                    sourceByteArray,
                    out int    imageWidth,
                    out int    imageHeight,
                    out bool   hasAlpha,
                    out bool   hasAnimation,
                    out string format
                );

                if(hasAlpha)
                {
                    targetBitmap = new Bitmap(imageWidth, imageHeight, PixelFormat.Format32bppArgb);
                }
                else
                {
                    targetBitmap = new Bitmap(imageWidth, imageHeight, PixelFormat.Format24bppRgb);
                }

                targetBitmapData = targetBitmap.LockBits
                (
                    new Rectangle(0, 0, imageWidth, imageHeight),
                    ImageLockMode.WriteOnly,
                    targetBitmap.PixelFormat
                );

                int    targetSize = targetBitmapData.Stride * imageHeight;
                IntPtr dataHandle = sourceHandle.AddrOfPinnedObject();

                if(targetBitmap.PixelFormat == PixelFormat.Format24bppRgb)
                {
                    WIN32Helper.WebPDecodeBGRInto
                    (
                        dataHandle,
                        sourceByteArray.Length,
                        targetBitmapData.Scan0,
                        targetSize,
                        targetBitmapData.Stride
                    );
                }
                else
                {
                    WIN32Helper.WEBPDecodeBGRAInto
                    (
                        dataHandle,
                        sourceByteArray.Length,
                        targetBitmapData.Scan0,
                        targetSize,
                        targetBitmapData.Stride
                    );
                }

                return targetBitmap;
            }
            finally
            {
                if(targetBitmapData != null)
                {
                    targetBitmap.UnlockBits(targetBitmapData);
                }

                if(sourceHandle.IsAllocated)
                {
                    sourceHandle.Free();
                }
            }
        }

        #endregion
        #region 디코딩하기 - Decode(sourceByteArray, options, pixelFormat)

        /// <summary>
        /// 디코딩하기
        /// </summary>
        /// <param name="sourceByteArray">소스 바이트 배열</param>
        /// <param name="options">WEBP 디코더 옵션</param>
        /// <param name="pixelFormat">픽셀 포맷</param>
        /// <returns>비트맵</returns>
        public Bitmap Decode(byte[] sourceByteArray, WEBPDecoderOptions options, PixelFormat pixelFormat = PixelFormat.DontCare)
        {
            GCHandle      sourceHandle     = GCHandle.Alloc(sourceByteArray, GCHandleType.Pinned);
            Bitmap        targetBitmap     = null;
            BitmapData    targetBitmapData = null;
            VP8StatusCode result;

            try
            {
                WEBPDecoderConfiguration configuration = new WEBPDecoderConfiguration();

                if(WIN32Helper.WEBPInitializeDecoderConfiguration(ref configuration) == 0)
                {
                    throw new Exception("WebPInitDecoderConfig failed. Wrong version?");
                }

                IntPtr dataHandle = sourceHandle.AddrOfPinnedObject();
                int    width;
                int    height;

                if(options.UseScaling == 0)
                {
                    result = WIN32Helper.WEBPGetFeatures(dataHandle, sourceByteArray.Length, ref configuration.Input);

                    if(result != VP8StatusCode.VP8_STATUS_OK)
                    {
                        throw new Exception("Failed WebPGetFeatures with error " + result);
                    }

                    if(options.UseCropping == 1)
                    {
                        if
                        (
                            options.CropLeft + options.CropWidth  > configuration.Input.Width ||
                            options.CropTop  + options.CropHeight > configuration.Input.Height
                        )
                        {
                            throw new Exception("Crop options exceeded WebP image dimensions");
                        }

                        width  = options.CropWidth;
                        height = options.CropHeight;
                    }
                }
                else
                {
                    width = options.ScaledWidth;
                    height = options.ScaledHeight;
                }

                configuration.Options.BypassFiltering        = options.BypassFiltering;
                configuration.Options.NoFancyUpsampling      = options.NoFancyUpsampling;
                configuration.Options.UseCropping            = options.UseCropping;
                configuration.Options.CropLeft               = options.CropLeft;
                configuration.Options.CropTop                = options.CropTop;
                configuration.Options.CropWidth              = options.CropWidth;
                configuration.Options.CropHeight             = options.CropHeight;
                configuration.Options.UseScaling             = options.UseScaling;
                configuration.Options.ScaledWidth            = options.ScaledWidth;
                configuration.Options.ScaledHeight           = options.ScaledHeight;
                configuration.Options.UseThreads             = options.UseThreads;
                configuration.Options.DitheringStrength      = options.DitheringStrength;
                configuration.Options.Flip                   = options.Flip;
                configuration.Options.AlphaDitheringStrength = options.AlphaDitheringStrength;

                if(configuration.Input.HasAlpha == 1)
                {
                    configuration.Output.ColorSpace = WEBP_CSP_MODE.MODE_bgrA;

                    targetBitmap = new Bitmap(configuration.Input.Width, configuration.Input.Height, PixelFormat.Format32bppArgb);
                }
                else
                {
                    configuration.Output.ColorSpace = WEBP_CSP_MODE.MODE_BGR;

                    targetBitmap = new Bitmap(configuration.Input.Width, configuration.Input.Height, PixelFormat.Format24bppRgb);
                }

                targetBitmapData = targetBitmap.LockBits
                (
                    new Rectangle(0, 0, targetBitmap.Width, targetBitmap.Height),
                    ImageLockMode.WriteOnly,
                    targetBitmap.PixelFormat
                );

                configuration.Output.U.RGBA.RGBA      = targetBitmapData.Scan0;
                configuration.Output.U.RGBA.Stride    = targetBitmapData.Stride;
                configuration.Output.U.RGBA.Size      = (UIntPtr)(targetBitmap.Height * targetBitmapData.Stride);
                configuration.Output.Height           = targetBitmap.Height;
                configuration.Output.Width            = targetBitmap.Width;
                configuration.Output.IsExternalMemory = 1;

                result = WIN32Helper.WEBPDecode(dataHandle, sourceByteArray.Length, ref configuration);

                if(result != VP8StatusCode.VP8_STATUS_OK)
                {
                    throw new Exception("Failed WebPDecode with error " + result);
                }

                WIN32Helper.WEBPFreeDecoderBuffer(ref configuration.Output);

                return targetBitmap;
            }
            finally
            {
                if(targetBitmapData != null)
                {
                    targetBitmap.UnlockBits(targetBitmapData);
                }

                if(sourceHandle.IsAllocated)
                {
                    sourceHandle.Free();
                }
            }
        }

        #endregion
        #region 로드하기 - Load(filePath)

        /// <summary>
        /// 로드하기
        /// </summary>
        /// <param name="filePath">파일 경로</param>
        /// <returns>비트맵</returns>
        public Bitmap Load(string filePath)
        {
            byte[] sourceByteArray = File.ReadAllBytes(filePath);

            return Decode(sourceByteArray);
        }

        #endregion
        #region 고속/저품질 썸네일 구하기 - GetThumbnailWithFasterLowQuality(sourceByteArray, width, height)

        /// <summary>
        /// 고속/저품질 썸네일 구하기
        /// </summary>
        /// <param name="sourceByteArray">소스 바이트 배열</param>
        /// <param name="width">너비</param>
        /// <param name="height">높이</param>
        /// <returns>비트맵</returns>
        public Bitmap GetThumbnailWithFasterLowQuality(byte[] sourceByteArray, int width, int height)
        {
            GCHandle   sourceHandle = GCHandle.Alloc(sourceByteArray, GCHandleType.Pinned);
            Bitmap     bitmap       = null;
            BitmapData bitmapData   = null;

            try
            {
                WEBPDecoderConfiguration configuration = new WEBPDecoderConfiguration();

                if(WIN32Helper.WEBPInitializeDecoderConfiguration(ref configuration) == 0)
                {
                    throw new Exception("WebPInitDecoderConfig failed. Wrong version?");
                }

                configuration.Options.BypassFiltering   = 1;
                configuration.Options.NoFancyUpsampling = 1;
                configuration.Options.UseThreads        = 1;
                configuration.Options.UseScaling        = 1;
                configuration.Options.ScaledWidth       = width;
                configuration.Options.ScaledHeight      = height;

                bitmap = new Bitmap(width, height, PixelFormat.Format24bppRgb);

                bitmapData = bitmap.LockBits(new Rectangle(0, 0, width, height), ImageLockMode.WriteOnly, bitmap.PixelFormat);

                configuration.Output.ColorSpace       = WEBP_CSP_MODE.MODE_BGR;
                configuration.Output.U.RGBA.RGBA      = bitmapData.Scan0;
                configuration.Output.U.RGBA.Stride    = bitmapData.Stride;
                configuration.Output.U.RGBA.Size      = (UIntPtr)(height * bitmapData.Stride);
                configuration.Output.Height           = height;
                configuration.Output.Width            = width;
                configuration.Output.IsExternalMemory = 1;

                IntPtr pinnedSourceHandle = sourceHandle.AddrOfPinnedObject();

                VP8StatusCode statusCode = WIN32Helper.WEBPDecode(pinnedSourceHandle, sourceByteArray.Length, ref configuration);

                if(statusCode != VP8StatusCode.VP8_STATUS_OK)
                {
                    throw new Exception("Failed WebPDecode with error " + statusCode);
                }

                WIN32Helper.WEBPFreeDecoderBuffer(ref configuration.Output);

                return bitmap;
            }
            finally
            {
                if(bitmapData != null)
                {
                    bitmap.UnlockBits(bitmapData);
                }

                if(sourceHandle.IsAllocated)
                {
                    sourceHandle.Free();
                }
            }
        }

        #endregion
        #region 저속/고품질 썸네일 구하기 - GetThumbnailWithSlowHighQuality(sourceByteArray, width, height)

        /// <summary>
        /// 저속/고품질 썸네일 구하기
        /// </summary>
        /// <param name="sourceByteArray">소스 바이트 배열</param>
        /// <param name="width">너비</param>
        /// <param name="height">높이</param>
        /// <returns>비트맵</returns>
        public Bitmap GetThumbnailWithSlowHighQuality(byte[] sourceByteArray, int width, int height)
        {
            GCHandle   sourceHandle = GCHandle.Alloc(sourceByteArray, GCHandleType.Pinned);
            Bitmap     bitmap       = null;
            BitmapData bitmapData   = null;

            try
            {
                WEBPDecoderConfiguration configuration = new WEBPDecoderConfiguration();

                if(WIN32Helper.WEBPInitializeDecoderConfiguration(ref configuration) == 0)
                {
                    throw new Exception("WebPInitDecoderConfig failed. Wrong version?");
                }

                IntPtr pinnedSourceHandle = sourceHandle.AddrOfPinnedObject();

                VP8StatusCode statusCode = WIN32Helper.WEBPGetFeatures(pinnedSourceHandle, sourceByteArray.Length, ref configuration.Input);

                if(statusCode != VP8StatusCode.VP8_STATUS_OK)
                {
                    throw new Exception("Failed WebPGetFeatures with error " + statusCode);
                }

                configuration.Options.BypassFiltering   = 0;
                configuration.Options.NoFancyUpsampling = 0;
                configuration.Options.UseThreads        = 1;
                configuration.Options.UseScaling        = 1;
                configuration.Options.ScaledWidth       = width;
                configuration.Options.ScaledHeight      = height;

                if(configuration.Input.HasAlpha == 1)
                {
                    configuration.Output.ColorSpace = WEBP_CSP_MODE.MODE_bgrA;

                    bitmap = new Bitmap(width, height, PixelFormat.Format32bppArgb);
                }
                else
                {
                    configuration.Output.ColorSpace = WEBP_CSP_MODE.MODE_BGR;

                    bitmap = new Bitmap(width, height, PixelFormat.Format24bppRgb);
                }

                bitmapData = bitmap.LockBits(new Rectangle(0, 0, width, height), ImageLockMode.WriteOnly, bitmap.PixelFormat);

                configuration.Output.U.RGBA.RGBA      = bitmapData.Scan0;
                configuration.Output.U.RGBA.Stride    = bitmapData.Stride;
                configuration.Output.U.RGBA.Size      = (UIntPtr)(height * bitmapData.Stride);
                configuration.Output.Height           = height;
                configuration.Output.Width            = width;
                configuration.Output.IsExternalMemory = 1;

                statusCode = WIN32Helper.WEBPDecode(pinnedSourceHandle, sourceByteArray.Length, ref configuration);

                if(statusCode != VP8StatusCode.VP8_STATUS_OK)
                {
                    throw new Exception("Failed WebPDecode with error " + statusCode);
                }

                WIN32Helper.WEBPFreeDecoderBuffer(ref configuration.Output);

                return bitmap;
            }
            finally
            {
                if(bitmapData != null)
                {
                    bitmap.UnlockBits(bitmapData);
                }

                if(sourceHandle.IsAllocated)
                {
                    sourceHandle.Free();
                }
            }
        }

        #endregion
        #region 저장하기 - Save(bitmap, filePath, quality)

        /// <summary>
        /// 저장하기
        /// </summary>
        /// <param name="bitmap">비트맵</param>
        /// <param name="filePath">파일 경로</param>
        /// <param name="quality">품질</param>
        public void Save(Bitmap bitmap, string filePath, int quality = 75)
        {
            byte[] sourceByteArray;

            sourceByteArray = EncodeLossy(bitmap, quality);

            File.WriteAllBytes(filePath, sourceByteArray);
        }

        #endregion
        #region 손실 인코딩하기 - EncodeLossy(bitmap, quality)

        /// <summary>
        /// 손실 인코딩하기
        /// </summary>
        /// <param name="bitmap">비트맵</param>
        /// <param name="quality">품질</param>
        /// <returns>바이트 배열</returns>
        public byte[] EncodeLossy(Bitmap bitmap, int quality = 75)
        {
            if(bitmap.Width == 0 || bitmap.Height == 0)
            {
                throw new ArgumentException("Bitmap contains no data.", "bmp");
            }

            if(bitmap.Width > WEBP_MAXIMUM_DIMENSION || bitmap.Height > WEBP_MAXIMUM_DIMENSION)
            {
                throw new NotSupportedException
                (
                    "Bitmap's dimension is too large. Max is " + WEBP_MAXIMUM_DIMENSION + "x" + WEBP_MAXIMUM_DIMENSION + " pixels."
                );
            }

            if(bitmap.PixelFormat != PixelFormat.Format24bppRgb && bitmap.PixelFormat != PixelFormat.Format32bppArgb)
            {
                throw new NotSupportedException("Only support Format24bppRgb and Format32bppArgb pixelFormat.");
            }

            BitmapData bitmapData          = null;
            IntPtr     unmanagedDataHandle = IntPtr.Zero;

            try
            {
                int size;
                
                bitmapData = bitmap.LockBits
                (
                    new Rectangle(0, 0, bitmap.Width, bitmap.Height),
                    ImageLockMode.ReadOnly,
                    bitmap.PixelFormat
                );

                if(bitmap.PixelFormat == PixelFormat.Format24bppRgb)
                {
                    size = WIN32Helper.WEBPEncodeBGR
                    (
                        bitmapData.Scan0,
                        bitmap.Width,
                        bitmap.Height,
                        bitmapData.Stride,
                        quality,
                        out unmanagedDataHandle
                    );
                }
                else
                {
                    size = WIN32Helper.WEBPEncodeBGRA
                    (
                        bitmapData.Scan0,
                        bitmap.Width,
                        bitmap.Height,
                        bitmapData.Stride,
                        quality,
                        out unmanagedDataHandle
                    );
                }

                if(size == 0)
                {
                    throw new Exception("Can´t encode WebP");
                }

                byte[] targetByteArray = new byte[size];

                Marshal.Copy(unmanagedDataHandle, targetByteArray, 0, size);

                return targetByteArray;
            }
            finally
            {
                if(bitmapData != null)
                {
                    bitmap.UnlockBits(bitmapData);
                }

                if(unmanagedDataHandle != IntPtr.Zero)
                {
                    WIN32Helper.WEBPFree(unmanagedDataHandle);
                }
            }
        }

        #endregion
        #region 손실 인코딩하기 - EncodeLossy(bitmap, quality, speed, information)

        /// <summary>
        /// 손실 인코딩하기
        /// </summary>
        /// <param name="bitmap">비트맵</param>
        /// <param name="quality">품질</param>
        /// <param name="speed">속도</param>
        /// <param name="information">정보 여부</param>
        /// <returns>바이트 배열</returns>
        public byte[] EncodeLossy(Bitmap bitmap, int quality, int speed, bool information = false)
        {
            WEBPConfiguration configuration = new WEBPConfiguration();

            if(WIN32Helper.WEBPInitializeConfiguration(ref configuration, WEBPPreset.WEBP_PRESET_DEFAULT, 75) == 0)
            {
                throw new Exception("Can´t configure preset");
            }

            configuration.Method = speed;

            if(configuration.Method > 6)
            {
                configuration.Method = 6;
            }

            configuration.Quality        = quality;
            configuration.AutoFilter     = 1;
            configuration.Pass           = speed + 1;
            configuration.SegmentCount   = 4;
            configuration.Partitions     = 3;
            configuration.ThreadLevel    = 1;
            configuration.AlphaQuality   = quality;
            configuration.AlphaFiltering = 2;
            configuration.UseSharpYUV    = 1;

            if(WIN32Helper.WEBPGetDecoderVersion() > 1082)
            {
                configuration.Preprocessing = 4;
                configuration.UseSharpYUV   = 1;
            }
            else
            {
                configuration.Preprocessing = 3;
            }

            return AdvancedEncode(bitmap, configuration, information);
        }

        #endregion
        #region 무손실 인코딩하기 - EncodeLossless(bitmap)

        /// <summary>
        /// 무손실 인코딩하기
        /// </summary>
        /// <param name="bitmap">비트맵</param>
        /// <returns>바이트 배열</returns>
        public byte[] EncodeLossless(Bitmap bitmap)
        {
            if(bitmap.Width == 0 || bitmap.Height == 0)
            {
                throw new ArgumentException("Bitmap contains no data.", "bmp");
            }

            if(bitmap.Width > WEBP_MAXIMUM_DIMENSION || bitmap.Height > WEBP_MAXIMUM_DIMENSION)
            {
                throw new NotSupportedException
                (
                    "Bitmap's dimension is too large. Max is " + WEBP_MAXIMUM_DIMENSION + "x" + WEBP_MAXIMUM_DIMENSION + " pixels."
                );
            }

            if(bitmap.PixelFormat != PixelFormat.Format24bppRgb && bitmap.PixelFormat != PixelFormat.Format32bppArgb)
            {
                throw new NotSupportedException("Only support Format24bppRgb and Format32bppArgb pixelFormat.");
            }

            BitmapData bitmapData          = null;
            IntPtr     unmanagedDataHandle = IntPtr.Zero;

            try
            {
                bitmapData = bitmap.LockBits
                (
                    new Rectangle(0, 0, bitmap.Width, bitmap.Height),
                    ImageLockMode.ReadOnly,
                    bitmap.PixelFormat
                );

                int size;

                if(bitmap.PixelFormat == PixelFormat.Format24bppRgb)
                {
                    size = WIN32Helper.WEBPEncodeLosslessBGR
                    (
                        bitmapData.Scan0,
                        bitmap.Width,
                        bitmap.Height,
                        bitmapData.Stride,
                        out unmanagedDataHandle
                    );
                }
                else
                {
                    size = WIN32Helper.WEBPEncodeLosslessBGRA
                    (
                        bitmapData.Scan0,
                        bitmap.Width,
                        bitmap.Height,
                        bitmapData.Stride,
                        out unmanagedDataHandle
                    );
                }

                byte[] targetByteArray = new byte[size];

                Marshal.Copy(unmanagedDataHandle, targetByteArray, 0, size);

                return targetByteArray;
            }
            finally
            {
                if(bitmapData != null)
                {
                    bitmap.UnlockBits(bitmapData);
                }

                if(unmanagedDataHandle != IntPtr.Zero)
                {
                    WIN32Helper.WEBPFree(unmanagedDataHandle);
                }
            }
        }

        #endregion
        #region 무손실 인코딩하기 - EncodeLossless(bitmap, speed)

        /// <summary>
        /// 무손실 인코딩하기
        /// </summary>
        /// <param name="bitmap">비트맵</param>
        /// <param name="speed">속도</param>
        /// <returns>바이트 배열</returns>
        public byte[] EncodeLossless(Bitmap bitmap, int speed)
        {
            WEBPConfiguration configuration = new WEBPConfiguration();

            if(WIN32Helper.WEBPInitializeConfiguration(ref configuration, WEBPPreset.WEBP_PRESET_DEFAULT, (speed + 1) * 10) == 0)
            {
                throw new Exception("Can´t config preset");
            }
                
            if(WIN32Helper.WEBPGetDecoderVersion() > 1082)
            {
                if(WIN32Helper.WEBPConfigLosslessPreset(ref configuration, speed) == 0)
                {
                    throw new Exception("Can´t configure lossless preset");
                }
            }
            else
            {
                configuration.Lossless = 1;
                configuration.Method   = speed;

                if(configuration.Method > 6)
                {
                    configuration.Method = 6;
                }

                configuration.Quality  = (speed + 1) * 10;
            }

            configuration.Pass           = speed + 1;
            configuration.ThreadLevel    = 1;
            configuration.AlphaFiltering = 2;
            configuration.UseSharpYUV    = 1;
            configuration.Exact          = 0;

            return AdvancedEncode(bitmap, configuration, false);
        }

        #endregion
        #region 무손실 근접 인코딩하기 - EncodeNearLossless(bitmap, quality, speed)

        /// <summary>
        /// 무손실 근접 인코딩하기
        /// </summary>
        /// <param name="bitmap">비트맵</param>
        /// <param name="quality">품질</param>
        /// <param name="speed">속도</param>
        /// <returns>바이트 배열</returns>
        public byte[] EncodeNearLossless(Bitmap bitmap, int quality, int speed = 9)
        {
            if(WIN32Helper.WEBPGetDecoderVersion() <= 1082)
            {
                throw new Exception("This DLL version not support EncodeNearLossless");
            }

            WEBPConfiguration configuration = new WEBPConfiguration();

            if(WIN32Helper.WEBPInitializeConfiguration(ref configuration, WEBPPreset.WEBP_PRESET_DEFAULT, (speed + 1) * 10) == 0)
            {
                throw new Exception("Can´t configure preset");
            }

            if(WIN32Helper.WEBPConfigLosslessPreset(ref configuration, speed) == 0)
            {
                throw new Exception("Can´t configure lossless preset");
            }

            configuration.Pass           = speed + 1;
            configuration.NearLossless   = quality;
            configuration.ThreadLevel    = 1;
            configuration.AlphaFiltering = 2;
            configuration.UseSharpYUV    = 1;
            configuration.Exact          = 0;

            return AdvancedEncode(bitmap, configuration, false);
        }

        #endregion
        #region 버전 구하기 - GetVersion()

        /// <summary>
        /// 버전 구하기
        /// </summary>
        /// <returns>버전</returns>
        public string GetVersion()
        {
            uint version      = (uint)WIN32Helper.WEBPGetDecoderVersion();
            uint revision     = version % 256;
            uint minorVersion = (version >> 8) % 256;
            uint majorVersion = (version >> 16) % 256;

            return majorVersion + "." + minorVersion + "." + revision;
        }

        #endregion
        #region 픽처 왜곡 구하기 - GetPictureDistortion(sourceBitmap, targetBitmap, metricType)

        /// <summary>
        /// 픽처 왜곡 구하기
        /// </summary>
        /// <param name="sourceBitmap">소스 비트맵</param>
        /// <param name="targetBitmap">타겟 비트맵</param>
        /// <param name="metricType">미터법 타입</param>
        /// <returns>픽처 왜곡 배열</returns>
        /// <remarks>
        /// 두 사진 간의 PSNR, SSIM 또는 LSIM 왜곡 메트릭을 계산한다.
        /// 경고 : 이 기능은 CPU를 많이 사용한다.
        /// </remarks>
        public float[] GetPictureDistortion(Bitmap sourceBitmap, Bitmap targetBitmap, int metricType)
        {
            WEBPPicture sourcePicture    = new WEBPPicture();
            WEBPPicture targetPicture    = new WEBPPicture();
            BitmapData  sourceBitmapData = null;
            BitmapData  targetBitmapData = null;
            float[]     targetArray      = new float[5];

            GCHandle pinnedTargetHandle = GCHandle.Alloc(targetArray, GCHandleType.Pinned);

            try
            {
                if(sourceBitmap == null)
                {
                    throw new Exception("Source picture is void");
                }

                if(targetBitmap == null)
                {
                    throw new Exception("Reference picture is void");
                }

                if(metricType > 2)
                {
                    throw new Exception("Bad metric_type. Use 0 = PSNR, 1 = SSIM, 2 = LSIM");
                }

                if(sourceBitmap.Width != targetBitmap.Width || sourceBitmap.Height != targetBitmap.Height)
                {
                    throw new Exception("Source and Reference pictures have different dimensions");
                }

                sourceBitmapData = sourceBitmap.LockBits
                (
                    new Rectangle(0, 0, sourceBitmap.Width, sourceBitmap.Height),
                    ImageLockMode.ReadOnly,
                    sourceBitmap.PixelFormat
                );

                sourcePicture = new WEBPPicture();

                if(WIN32Helper.WEBPInitializePictureInternal(ref sourcePicture) != 1)
                {
                    throw new Exception("Can´t initialize WebPPictureInit");
                }

                sourcePicture.Width  = (int)sourceBitmap.Width;
                sourcePicture.Height = (int)sourceBitmap.Height;

                if(sourceBitmapData.PixelFormat == PixelFormat.Format32bppArgb)
                {
                    sourcePicture.UseARGB = 1;

                    if(WIN32Helper.WEBPPictureImportBGRA(ref sourcePicture, sourceBitmapData.Scan0, sourceBitmapData.Stride) != 1)
                    {
                        throw new Exception("Can´t allocate memory in WebPPictureImportBGR");
                    }
                }
                else
                {
                    sourcePicture.UseARGB = 0;

                    if(WIN32Helper.WEBPPictureImportBGR(ref sourcePicture, sourceBitmapData.Scan0, sourceBitmapData.Stride) != 1)
                    {
                        throw new Exception("Can´t allocate memory in WebPPictureImportBGR");
                    }
                }

                targetBitmapData = targetBitmap.LockBits
                (
                    new Rectangle(0, 0, targetBitmap.Width, targetBitmap.Height),
                    ImageLockMode.ReadOnly,
                    targetBitmap.PixelFormat
                );

                targetPicture = new WEBPPicture();

                if(WIN32Helper.WEBPInitializePictureInternal(ref targetPicture) != 1)
                {
                    throw new Exception("Can´t initialize WebPPictureInit");
                }

                targetPicture.Width   = (int)targetBitmap.Width;
                targetPicture.Height  = (int)targetBitmap.Height;
                targetPicture.UseARGB = 1;

                if(sourceBitmapData.PixelFormat == PixelFormat.Format32bppArgb)
                {
                    sourcePicture.UseARGB = 1;

                    if(WIN32Helper.WEBPPictureImportBGRA(ref targetPicture, targetBitmapData.Scan0, targetBitmapData.Stride) != 1)
                    {
                        throw new Exception("Can´t allocate memory in WebPPictureImportBGR");
                    }
                }
                else
                {
                    sourcePicture.UseARGB = 0;

                    if(WIN32Helper.WEBPPictureImportBGR(ref targetPicture, targetBitmapData.Scan0, targetBitmapData.Stride) != 1)
                    {
                        throw new Exception("Can´t allocate memory in WebPPictureImportBGR");
                    }
                }

                IntPtr targetHandle = pinnedTargetHandle.AddrOfPinnedObject();

                if(WIN32Helper.WEBPPictureDistortion(ref sourcePicture, ref targetPicture, metricType, targetHandle) != 1)
                {
                    throw new Exception("Can´t measure.");
                }

                return targetArray;
            }
            finally
            {
                if(sourceBitmapData != null)
                {
                    sourceBitmap.UnlockBits(sourceBitmapData);
                }

                if(targetBitmapData != null)
                {
                    targetBitmap.UnlockBits(targetBitmapData);
                }

                if(sourcePicture.ARGB != IntPtr.Zero)
                {
                    WIN32Helper.WEBPPictureFree(ref sourcePicture);
                }

                if(targetPicture.ARGB != IntPtr.Zero)
                {
                    WIN32Helper.WEBPPictureFree(ref targetPicture);
                }

                if(pinnedTargetHandle.IsAllocated)
                {
                    pinnedTargetHandle.Free();
                }
            }
        }

        #endregion

        #region 리소스 해제하기 - Dispose()

        /// <summary>
        /// 리소스 해제하기
        /// </summary>
        public void Dispose()
        {
            GC.SuppressFinalize(this);
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 메모리 라이터 - MemoryWriter(dataHandle, dataSize, picture)

        /// <summary>
        /// 메모리 라이터
        /// </summary>
        /// <param name="dataHandle">데이터 핸들</param>
        /// <param name="dataSize">데이터 크기</param>
        /// <param name="picture">WEBP 픽처</param>
        /// <returns>처리 결과</returns>
        private int MemoryWriter([InAttribute()] IntPtr dataHandle, UIntPtr dataSize, ref WEBPPicture picture)
        {
            WIN32Helper.CopyMemory(picture.CustomPointer, dataHandle, (uint)dataSize);

            picture.CustomPointer = new IntPtr(picture.CustomPointer.ToInt64() + (int)dataSize);

            return 1;
        }
        
        #endregion
        #region 개선 인코딩하기 - AdvancedEncode(bitmap, configuration, information)

        /// <summary>
        /// 개선 인코딩하기
        /// </summary>
        /// <param name="bitmap">비트맵</param>
        /// <param name="configuration">WEBP 구성</param>
        /// <param name="information">정보 여부</param>
        /// <returns>바이트 배열</returns>
        private byte[] AdvancedEncode(Bitmap bitmap, WEBPConfiguration configuration, bool information)
        {
            byte[]                  targetByteArray = null;
            byte[]                  sourceByteArray = null;
            WEBPPicture             picture           = new WEBPPicture();
            BitmapData              bitmapData        = null;
            WEBPAuxiliaryStatistics statistics        = new WEBPAuxiliaryStatistics();
            IntPtr                  statisticsHandle  = IntPtr.Zero;
            GCHandle                pinnedArrayHandle = new GCHandle();
            int                     sourceSize;

            try
            {
                if(WIN32Helper.WEBPValidateConfiguration(ref configuration) != 1)
                {
                    throw new Exception("Bad configuration parameters");
                }

                if(bitmap.Width == 0 || bitmap.Height == 0)
                {
                    throw new ArgumentException("Bitmap contains no data.", "bmp");
                }

                if(bitmap.Width > WEBP_MAXIMUM_DIMENSION || bitmap.Height > WEBP_MAXIMUM_DIMENSION)
                {
                    throw new NotSupportedException
                    (
                        "Bitmap's dimension is too large. Max is " + WEBP_MAXIMUM_DIMENSION + "x" + WEBP_MAXIMUM_DIMENSION + " pixels."
                    );
                }

                if(bitmap.PixelFormat != PixelFormat.Format24bppRgb && bitmap.PixelFormat != PixelFormat.Format32bppArgb)
                {
                    throw new NotSupportedException("Only support Format24bppRgb and Format32bppArgb pixelFormat.");
                }

                bitmapData = bitmap.LockBits
                (
                    new Rectangle(0, 0, bitmap.Width, bitmap.Height),
                    ImageLockMode.ReadOnly,
                    bitmap.PixelFormat
                );

                if(WIN32Helper.WEBPInitializePictureInternal(ref picture) != 1)
                {
                    throw new Exception("Can´t initialize WebPPictureInit");
                }

                picture.Width   = (int)bitmap.Width;
                picture.Height  = (int)bitmap.Height;
                picture.UseARGB = 1;

                if(bitmap.PixelFormat == PixelFormat.Format32bppArgb)
                {
                    int result = WIN32Helper.WEBPPictureImportBGRA(ref picture, bitmapData.Scan0, bitmapData.Stride);

                    if(result != 1)
                    {
                        throw new Exception("Can´t allocate memory in WebPPictureImportBGRA");
                    }

                    picture.ColorSpace = (uint)WEBP_CSP_MODE.MODE_bgrA;

                    sourceSize = bitmap.Width * bitmap.Height * 32;
                    sourceByteArray = new byte[bitmap.Width * bitmap.Height * 32]; 
                }
                else
                {
                    int result = WIN32Helper.WEBPPictureImportBGR(ref picture, bitmapData.Scan0, bitmapData.Stride);

                    if(result != 1)
                    {
                        throw new Exception("Can´t allocate memory in WebPPictureImportBGR");
                    }

                    sourceSize = bitmap.Width * bitmap.Height * 24;
                }

                if(information)
                {
                    statistics       = new WEBPAuxiliaryStatistics();
                    statisticsHandle = Marshal.AllocHGlobal(Marshal.SizeOf(statistics));

                    Marshal.StructureToPtr(statistics, statisticsHandle, false);

                    picture.Statistics = statisticsHandle;
                }

                if(sourceSize > 2147483591)
                {
                    sourceSize = 2147483591;
                }

                sourceByteArray = new byte[bitmap.Width * bitmap.Height * 32];

                pinnedArrayHandle = GCHandle.Alloc(sourceByteArray, GCHandleType.Pinned);

                IntPtr initialHandle = pinnedArrayHandle.AddrOfPinnedObject();

                picture.CustomPointer = initialHandle;

                WIN32Helper.memoryWriteDelegate = new WIN32Helper.WEBPMemoryWriteDelegate(MemoryWriter);

                picture.Writer = Marshal.GetFunctionPointerForDelegate(WIN32Helper.memoryWriteDelegate);

                if(WIN32Helper.WEBPEncode(ref configuration, ref picture) != 1)
                {
                    throw new Exception("Encoding error : " + ((WEBPEncodingError)picture.ErrorCode).ToString());
                }

                WIN32Helper.memoryWriteDelegate = null;

                bitmap.UnlockBits(bitmapData);

                bitmapData = null;

                int size = (int)((long)picture.CustomPointer - (long)initialHandle);

                targetByteArray = new byte[size];

                Array.Copy(sourceByteArray, targetByteArray, size);

                pinnedArrayHandle.Free();

                sourceByteArray = null;

                if(information)
                {
                    statistics = (WEBPAuxiliaryStatistics)Marshal.PtrToStructure(statisticsHandle, typeof(WEBPAuxiliaryStatistics));

                    MessageBox.Show
                    (
                        "Dimension      : " + picture.Width + " x " + picture.Height + " pixels\n"        +
                        "Output         : " + statistics.CodedSize                 + " bytes\n"           +
                        "PSNR Y         : " + statistics.PSNRY                     + " db\n"              +
                        "PSNR u         : " + statistics.PSNRU                     + " db\n"              +
                        "PSNR v         : " + statistics.PSNRV                     + " db\n"              +
                        "PSNR ALL       : " + statistics.PSNRALL                   + " db\n"              +
                        "Block intra4   : " + statistics.BlockCountIntra4          + "\n"                 +
                        "Block intra16  : " + statistics.BlockCountIntra16         + "\n"                 +
                        "Block skipped  : " + statistics.BlockCountSkipped         + "\n"                 +
                        "Header size    : " + statistics.HeaderByteCount           + " bytes\n"           +
                        "Mode-partition : " + statistics.ModePartition0            + " bytes\n"           +
                        "Macro-blocks 0 : " + statistics.SegmentSizeSegments0      + " residuals bytes\n" +
                        "Macro-blocks 1 : " + statistics.SegmentSizeSegments1      + " residuals bytes\n" +
                        "Macro-blocks 2 : " + statistics.SegmentSizeSegments2      + " residuals bytes\n" +
                        "Macro-blocks 3 : " + statistics.SegmentSizeSegments3      + " residuals bytes\n" +
                        "Quantizer    0 : " + statistics.SegmentQuantizerSegments0 + " residuals bytes\n" +
                        "Quantizer    1 : " + statistics.SegmentQuantizerSegments1 + " residuals bytes\n" +
                        "Quantizer    2 : " + statistics.SegmentQuantizerSegments2 + " residuals bytes\n" +
                        "Quantizer    3 : " + statistics.SegmentQuantizerSegments3 + " residuals bytes\n" +
                        "Filter level 0 : " + statistics.SegmentLevelSegments0     + " residuals bytes\n" +
                        "Filter level 1 : " + statistics.SegmentLevelSegments1     + " residuals bytes\n" +
                        "Filter level 2 : " + statistics.SegmentLevelSegments2     + " residuals bytes\n" +
                        "Filter level 3 : " + statistics.SegmentLevelSegments3     + " residuals bytes\n",
                        "Compression statistics"
                    );
                }

                return targetByteArray;
            }
            finally
            {
                if(pinnedArrayHandle.IsAllocated)
                {
                    pinnedArrayHandle.Free();
                }

                if(statisticsHandle != IntPtr.Zero)
                {
                    Marshal.FreeHGlobal(statisticsHandle);
                }

                if(bitmapData != null)
                {
                    bitmap.UnlockBits(bitmapData);
                }

                if(picture.ARGB != IntPtr.Zero)
                {
                    WIN32Helper.WEBPPictureFree(ref picture);
                }
            }
        }

        #endregion
    }
}